from typing import Any, Generic, Iterable, Optional, TypeAlias, TypeVar

from rainbowadn.atomic import *
from rainbowadn.collection.comparison import *
from rainbowadn.collection.keymetadata import *
from rainbowadn.collection.trees.binary import *
from rainbowadn.collection.trees.binary.core import *
from rainbowadn.core import *
from rainbowadn.flow.core import *
from rainbowadn.flow.primitive import *
from rainbowadn.flow.verification.core import *
from rainbowadn.inlining import *
from rainbowadn.nullability import *
from ._binaryflow import *
from ._flowtree import *

__all__ = ('FlowStandard', 'FlowStandardFactory',)

KeyT = TypeVar('KeyT')
FS: TypeAlias = 'FlowStandard[KeyT]'
ABT: TypeAlias = 'ActiveBinaryTree[KeyT, Integer]'
BP: TypeAlias = 'BinaryProtocolized[KeyT, Integer, ABT]'


class FlowStandard(
    RecursiveMentionable,
    FlowTree[HashPoint[KeyT], FS],
    Generic[KeyT]
):
    def points(self) -> Iterable[HashPoint]:
        return self.protocolized.tree.reference.points()

    def __bytes__(self):
        return bytes(self.protocolized.tree.reference)

    def __factory__(self) -> RainbowFactory['FlowStandard[KeyT]']:
        return FlowStandardFactory(
            self.protocolized.tree.reference.factory,
            self.protocolized.creation.comparator
        )

    def __init__(self, protocolized: BP):
        assert isinstance(protocolized, BinaryProtocolized)
        self.protocolized = protocolized

    async def contains(self, key: HashPoint[KeyT], *, exact: bool) -> bool:
        return await self.protocolized.tree.contains(key, exact=exact)

    def _protocolized(self: FS) -> BP:
        return self.protocolized

    @classmethod
    def _protocolized_mapper(cls) -> Mapper[FS, BP]:
        return CallableMapper(cls._protocolized)

    async def verify_subset(self, trees: Reducer[FS, CheckResult]) -> bool:
        assert isinstance(trees, Reducer)
        reducer: Reducer[BP, CheckResult] = MapReducer(self._protocolized_mapper(), trees)
        assert_true(await VerifySubsetAction(reducer).on(self.protocolized))
        return True

    async def verify_empty(self) -> bool:
        assert_true(
            await ReduceVerification(
                MapperVerification(ConstMapper(False))
            ).verify(
                await self.reducer()
            )
        )
        return True

    async def reducer(self) -> Reducer[HashPoint[KeyT], Any]:
        return BinaryReducer(self.protocolized).loose()

    def __eq__(self, other):
        if isinstance(other, FlowStandard):
            return self.protocolized.tree == other.protocolized.tree
        else:
            return NotImplemented

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        return await self.protocolized.tree.reference.str(tab)


class FlowStandardFactory(Inlining[FlowStandard[KeyT]], Generic[KeyT]):
    def __init__(
            self,
            factory: RainbowFactory[BinaryTree[KeyMetadata[KeyT, Integer]]],
            comparator: Comparator[KeyT]
    ):
        assert isinstance(factory, RainbowFactory)
        assert isinstance(comparator, Comparator)
        self.factory: RainbowFactory[
            NullableReference[BinaryTree[KeyMetadata[KeyT, Integer]]]
        ] = NullableReference.f(factory)
        self.comparator = comparator

    def size(self) -> Optional[int]:
        return Inlining.factory_size(self.factory)

    @classmethod
    def of(cls, factory: RainbowFactory[KeyT], comparator: Comparator[KeyT]) -> 'FlowStandardFactory[KeyT]':
        assert isinstance(factory, RainbowFactory)
        return FlowStandardFactory(
            BinaryTreeFactory(KeyMetadataFactory(factory, Integer.factory())),
            comparator
        )

    @classmethod
    async def off(
            cls, factory: RainbowFactory[KeyT], comparator: Comparator[KeyT], keys: Iterable[KeyT]
    ) -> 'FlowStandard[KeyT]':
        assert isinstance(factory, RainbowFactory)
        abt: ActiveBinaryTree[KeyT, Integer] = cls.empty(factory, comparator).protocolized.tree
        for key in keys:
            abt = await abt.add(HashPoint.of(key))
        return FlowStandard(abt.protocolized())

    @classmethod
    def protocol(cls, comparator: Comparator[KeyT]) -> BinaryBalancing[KeyT, Integer, ABT]:
        return AVL(comparator)

    @classmethod
    def empty(cls, factory: RainbowFactory[KeyT], comparator: Comparator[KeyT]) -> 'FlowStandard[KeyT]':
        assert isinstance(factory, RainbowFactory)
        return FlowStandard(
            ActiveBinaryTree.empty(cls.protocol(comparator), factory).protocolized()
        )

    def from_bytes(self, source: bytes, resolver: HashResolver) -> FlowStandard[KeyT]:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return FlowStandard(
            ActiveBinaryTree(
                self.protocol(self.comparator),
                self.factory.from_bytes(source, resolver)
            ).protocolized()
        )
