import heapq
from typing import Callable, Generic, Iterable, Optional, Type, TypeAlias, TypeVar, overload

from rainbowadn.core import *
from .inlining import *

__all__ = ('IAuto', 'Auto', 'IAutoFactory',)

_IList: TypeAlias = list[tuple[int, Mentionable]]
_UList: TypeAlias = list[tuple[int, HashPoint]]
_SList: TypeAlias = list[tuple[int, RainbowFactory, int]]
_VList: TypeAlias = list[tuple[int, RainbowFactory]]
_MCall: TypeAlias = Callable[['Auto', _IList, _UList], None]
_MTuple: TypeAlias = tuple[int, _MCall, int]
_IAuto = TypeVar('_IAuto')


class IAuto(RecursiveMentionable):
    def __init__(
            self,
            inlined: _IList,
            uninlined: _UList
    ):
        assert isinstance(inlined, list)
        assert isinstance(uninlined, list)
        for index, mentionable in inlined:
            assert isinstance(index, int)
            assert isinstance(mentionable, Mentionable)
        for index, hashpoint in uninlined:
            assert isinstance(index, int)
            assert isinstance(hashpoint, HashPoint)
        self.inlined = inlined
        self.uninlined = uninlined

    def points(self) -> Iterable[HashPoint]:
        inlined_points: list[tuple[int, HashPoint]] = []
        uninlined_points: list[tuple[int, HashPoint]] = []
        for index, inlined in self.inlined:
            if isinstance(inlined, RecursiveMentionable):
                for point in inlined.points():
                    assert isinstance(point, HashPoint)
                    inlined_points.append((index, point))
        for index, uninlined in self.uninlined:
            uninlined_points.append((index, uninlined))
        merged_points: Iterable[tuple[int, HashPoint]] = heapq.merge(inlined_points, uninlined_points)
        for _, point in merged_points:
            assert isinstance(point, HashPoint)
            yield point

    def __bytes__(self):
        inlined_bytes: list[tuple[int, bytes]] = []
        uninlined_bytes: list[tuple[int, bytes]] = []
        for index, inlined in self.inlined:
            inlined_bytes.append((index, bytes(inlined)))
        for index, uninlined in self.uninlined:
            uninlined_bytes.append((index, bytes(uninlined)))
        merged_bytes: Iterable[tuple[int, bytes]] = heapq.merge(inlined_bytes, uninlined_bytes)
        return b''.join(source for _, source in merged_bytes)

    def __factory__(self: _IAuto) -> RainbowFactory[_IAuto]:
        assert isinstance(self, IAuto)
        sized: _SList = []
        inlined_unsized: _VList = []
        uninlined_unsized: _VList = []
        for index, inlined in self.inlined:
            factory = inlined.__factory__()
            size = Inlining.factory_size(factory)
            if size is None:
                inlined_unsized.append((index, factory))
            else:
                assert isinstance(size, int)
                sized.append((index, factory, size))
        for index, uninlined in self.uninlined:
            uninlined_unsized.append((index, uninlined.factory))
        merged_unsized: Iterable[tuple[int, RainbowFactory]] = heapq.merge(inlined_unsized, uninlined_unsized)
        return IAutoFactory(
            type(self),
            sized,
            list(merged_unsized)
        )

    def hashpoints(self) -> Iterable[HashPoint]:
        inlined_hashpoints: list[tuple[int, HashPoint]] = []
        uninlined_hashpoints: list[tuple[int, HashPoint]] = []
        for index, inlined in self.inlined:
            inlined_hashpoints.append((index, HashPoint.of(inlined)))
        for index, uninlined in self.uninlined:
            uninlined_hashpoints.append((index, uninlined))
        merged_hashpoints: Iterable[tuple[int, HashPoint]] = heapq.merge(inlined_hashpoints, uninlined_hashpoints)
        for _, hashpoint in merged_hashpoints:
            yield hashpoint

    async def values(self) -> Iterable[Mentionable]:
        return await gather(*map(HashPoint.resolve, self.hashpoints()))

    @classmethod
    def auto_f(cls: Type[_IAuto], *factories: RainbowFactory) -> RainbowFactory[_IAuto]:
        assert issubclass(cls, IAuto)
        return IAutoFactory.of(cls, *factories)

    @classmethod
    def auto_of(cls: Type[_IAuto], *values: Mentionable) -> _IAuto:
        assert issubclass(cls, IAuto)
        sized: _IList = []
        unsized: _IList = []
        for index, value in enumerate(values):
            assert isinstance(value, Mentionable)
            factory = value.__factory__()
            if Inlining.factory_size(factory) is None:
                unsized.append((index, value))
            else:
                sized.append((index, value))
        inlined: _IList
        uninlined: _UList
        match unsized:
            case [(_, _)]:
                inlined = list(heapq.merge(sized, unsized))
                uninlined = []
            case _:
                inlined = sized
                uninlined = [(index, HashPoint.of(value)) for index, value in unsized]
        return cls(inlined, uninlined)

    @classmethod
    async def _utuple_to_ituple(cls, utuple: tuple[int, HashPoint]) -> tuple[int, Mentionable]:
        assert isinstance(utuple, tuple)
        index, hashpoint = utuple
        return index, await hashpoint.resolve()

    @classmethod
    async def auto_off(cls: Type[_IAuto], *hashpoints: HashPoint) -> _IAuto:
        assert issubclass(cls, IAuto)
        sized: _UList = []
        unsized: _UList = []
        for index, hashpoint in enumerate(hashpoints):
            assert isinstance(hashpoint, HashPoint)
            factory = hashpoint.factory
            if Inlining.factory_size(factory) is None:
                unsized.append((index, hashpoint))
            else:
                sized.append((index, hashpoint))
        inlined: _IList
        uninlined: _UList
        match unsized:
            case [(_, _)]:
                inlined = list(await gather(*map(cls._utuple_to_ituple, heapq.merge(sized, unsized))))
                uninlined = []
            case _:
                inlined = list(await gather(*map(cls._utuple_to_ituple, sized)))
                uninlined = unsized
        return cls(inlined, uninlined)

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        formatted = await gather(*(hash_point_format(hashpoint, tab) for hashpoint in self.hashpoints()))
        return f'{tabulate(tab)}'.join(formatted)


T = TypeVar('T')
T0 = TypeVar('T0')
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')


class Auto:
    def __init__(self, source: bytes, resolver: HashResolver):
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        self.source = source
        self.resolver = resolver

    def as_hashpoint(self, factory: RainbowFactory[T]) -> HashPoint[T]:
        assert isinstance(factory, RainbowFactory)
        assert_eq(len(self.source), HashPoint.HASH_LENGTH)
        return ResolverMetaOrigin(self.resolver).hash_point(factory, self.source)

    def as_value(self, factory: RainbowFactory[T]) -> T:
        return factory.from_bytes(self.source, self.resolver)

    def as_mentionable(self, factory: RainbowFactory) -> Mentionable:
        return self.as_value(factory)

    def sub(self, start: Optional[int], stop: Optional[int]) -> 'Auto':
        return Auto(self.source[start:stop], self.resolver)

    def __enter__(self):
        self.__index = 0

    def __exit__(self, exc_type, exc_val, exc_tb):
        assert_eq(len(self.source), self.__index)

    def next(self, size: int) -> 'Auto':
        assert isinstance(size, int)
        index = self.__index
        self.__index += size
        return self.sub(index, self.__index)

    @classmethod
    def _static_size(cls, *factories: RainbowFactory) -> int:
        return sum(Inlining.factory_size(factory) for factory in factories)

    def _static_simple(self, *factories: RainbowFactory) -> Iterable[Mentionable]:
        with self:
            for factory in factories:
                size = Inlining.factory_size(factory)
                assert isinstance(size, int)
                yield self.next(size).as_mentionable(factory)

    def _static_separate(self, infix_position: int, *factories: RainbowFactory) -> Iterable[Mentionable]:
        prefix = factories[:infix_position]
        postfix = factories[infix_position + 1:]
        prefix_size = self._static_size(*prefix)
        postfix_size = self._static_size(*postfix)
        postfix_start = len(self.source) - postfix_size
        yield from self.sub(None, prefix_size)._static_simple(*prefix)
        yield self.sub(prefix_size, postfix_start).as_mentionable(factories[infix_position])
        yield from self.sub(postfix_start, None)._static_simple(*postfix)

    def _static(self, *factories: RainbowFactory) -> Iterable[Mentionable]:
        infix_position: Optional[int] = None
        for index, factory in enumerate(factories):
            size: Optional[int] = Inlining.factory_size(factory)
            if size is None:
                if infix_position is None:
                    infix_position = index
                else:
                    raise ValueError('static auto parse does not allow uninlineable values (two or more unsized)')
        if infix_position is None:
            yield from self._static_simple(*factories)
        else:
            yield from self._static_separate(infix_position, *factories)

    @overload
    def static(
            self, t0: RainbowFactory[T0]
    ) -> tuple[T0]:
        ...

    @overload
    def static(
            self, t0: RainbowFactory[T0], t1: RainbowFactory[T1]
    ) -> tuple[T0, T1]:
        ...

    @overload
    def static(
            self, t0: RainbowFactory[T0], t1: RainbowFactory[T1], t2: RainbowFactory[T2]
    ) -> tuple[T0, T1, T2]:
        ...

    @overload
    def static(
            self, t0: RainbowFactory[T0], t1: RainbowFactory[T1], t2: RainbowFactory[T2], t3: RainbowFactory[T3]
    ) -> tuple[T0, T1, T2, T3]:
        ...

    @overload
    def static(
            self, t0: RainbowFactory[T0], t1: RainbowFactory[T1], t2: RainbowFactory[T2], t3: RainbowFactory[T3],
            t4: RainbowFactory[T4]
    ) -> tuple[T0, T1, T2, T3, T4]:
        ...

    def static(self, *factories: RainbowFactory) -> tuple[Mentionable, ...]:
        return tuple(self._static(*factories))


class IAutoFactory(Inlining[IAuto], Generic[_IAuto]):
    def __init__(
            self,
            typed: Type[_IAuto],
            sized: _SList,
            unsized: _VList
    ):
        assert issubclass(typed, IAuto)
        assert isinstance(sized, list)
        assert isinstance(unsized, list)
        self.typed = typed
        for index, factory, size in sized:
            assert isinstance(index, int)
            assert isinstance(factory, RainbowFactory)
            assert isinstance(size, int)
        for index, factory in unsized:
            assert isinstance(index, int)
            assert isinstance(factory, RainbowFactory)
        self.sized = sized
        self.unsized = unsized
        match unsized:
            case [(infix_index, infix_factory)]:
                self.infix = True
                self.infix_index: int = infix_index
                self.infix_factory: RainbowFactory = infix_factory
                self.prefix: _SList = list(
                    (index, factory, size) for index, factory, size in sized if index < infix_index
                )
                self.prefix_size = self._affix_size(self.prefix)
                assert isinstance(self.prefix_size, int)
                self.postfix: _SList = list(
                    (index, factory, size) for index, factory, size in sized if index > infix_index
                )
                self.postfix_size = self._affix_size(self.postfix)
                assert isinstance(self.postfix_size, int)
            case _:
                self.infix = False
                self.merged = self._merged()
                assert isinstance(self.merged, list)

    @classmethod
    def of(cls, typed: Type[_IAuto], *factories: RainbowFactory) -> Inlining[_IAuto]:
        sized: _SList = []
        unsized: _VList = []
        for index, factory in enumerate(factories):
            assert isinstance(factory, RainbowFactory)
            size = Inlining.factory_size(factory)
            if size is None:
                unsized.append((index, factory))
            else:
                assert isinstance(size, int)
                sized.append((index, factory, size))
        return cls(typed, sized, unsized)

    @classmethod
    def _affix_size(cls, affix: _SList) -> int:
        assert isinstance(affix, list)
        return sum(size for _, _, size in affix)

    @classmethod
    def _unsized_mtuple(cls, index: int, factory: RainbowFactory) -> _MTuple:
        assert isinstance(index, int)
        assert isinstance(factory, RainbowFactory)

        def wrap(auto: Auto, inlined: _IList, uninlined: _UList) -> None:
            assert isinstance(auto, Auto)
            assert isinstance(inlined, list)
            assert isinstance(uninlined, list)
            assert_eq(len(auto.source), HashPoint.HASH_LENGTH)
            uninlined.append((index, auto.as_hashpoint(factory)))

        return index, wrap, HashPoint.HASH_LENGTH

    @classmethod
    def _sized_mtuple(cls, index: int, factory: RainbowFactory, size: int) -> _MTuple:
        assert isinstance(index, int)
        assert isinstance(factory, RainbowFactory)
        assert isinstance(size, int)

        def wrap(auto: Auto, inlined: _IList, uninlined: _UList) -> None:
            assert isinstance(auto, Auto)
            assert isinstance(inlined, list)
            assert isinstance(uninlined, list)
            assert_eq(len(auto.source), size)
            inlined.append((index, auto.as_mentionable(factory)))

        return index, wrap, size

    def _merged(self) -> list[tuple[_MCall, int]]:
        unsized_mergeable: list[_MTuple] = []
        sized_mergeable: list[_MTuple] = []
        for index, factory in self.unsized:
            unsized_mergeable.append(
                self._unsized_mtuple(index, factory)
            )
        for index, factory, size in self.sized:
            sized_mergeable.append(
                self._sized_mtuple(index, factory, size)
            )
        merged: Iterable[_MTuple] = heapq.merge(unsized_mergeable, sized_mergeable)
        return [(method, size) for _, method, size in merged]

    def size(self) -> Optional[int]:
        if self.infix:
            return None
        else:
            return sum(size for _, size in self.merged)

    @classmethod
    def _parse_affix(
            cls, auto: Auto, affix: _SList
    ) -> Iterable[tuple[int, Mentionable]]:
        assert isinstance(auto, Auto)
        assert isinstance(affix, list)
        with auto:
            for index, factory, size in affix:
                yield index, auto.next(size).as_mentionable(factory)

    def _parse_infix(self, auto: Auto) -> Iterable[tuple[int, Mentionable]]:
        assert isinstance(auto, Auto)
        yield from self._parse_affix(auto.sub(None, self.prefix_size), self.prefix)
        postfix_start = len(auto.source) - self.postfix_size
        yield self.infix_index, auto.sub(self.prefix_size, postfix_start).as_mentionable(self.infix_factory)
        yield from self._parse_affix(auto.sub(postfix_start, None), self.postfix)

    def _parse_merged(self, auto: Auto) -> _IAuto:
        assert isinstance(auto, Auto)
        inlined: _IList = []
        uninlined: _UList = []
        with auto:
            for method, size in self.merged:
                method(auto.next(size), inlined, uninlined)
        return self.typed(inlined, uninlined)

    def from_bytes(self, source: bytes, resolver: HashResolver) -> _IAuto:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        auto = Auto(source, resolver)
        if self.infix:
            return self.typed(list(self._parse_infix(auto)), [])
        else:
            return self._parse_merged(auto)
